using System;
using System.Reflection;
using System.Diagnostics;
using System.Configuration;
using System.Web;
using System.Web.SessionState;

namespace PersistDemo
{
	/// <summary>
	/// Page persistence enum: save to Session or Application property collection, 
	///	or from web configuration file <c>appSettings</c> tag (restore only).
	/// </summary>
	public enum PagePersistEnum { 
		/// <summary>
		/// persist at the session level
		/// </summary>
		ePersistSession, 
		/// <summary>
		/// persist at the session level
		/// </summary>
		ePersistApplication, 
		/// <summary>
		/// persist at the session level
		/// </summary>
		ePersistWebConfig };
	/// <summary>
	/// Page persistence exception: thrown by worker functions of PagePersister.  The original
	///	exception is stored as the "inner" exception.
	/// </summary>
	public class PagePersistException : Exception
	{
		/// <summary>
		/// Ctor for page persistence exception object.
		/// </summary>
		/// <param name="sMsg">the error message</param>
		public PagePersistException (string sMsg)  
			: base(sMsg)
			{}
			
		/// <summary>
		/// Ctor for page persistence exception object.
		/// </summary>
		/// <param name="sMsg">the error message</param>
		/// <param name="exInner">the inner exception</param>
		public PagePersistException (string sMsg, Exception exInner)
			: base(sMsg,exInner)
			{}		
	}
	
	/// <summary>
	/// Extended attribute for decorating object-level fields with persistence information.
	/// </summary>
	[AttributeUsage(AttributeTargets.Field)]
	public class PagePersistAttribute : System.Attribute
	{
		/// <summary>
		/// Constructor for the attribute; takes a persistence medium enum: session, application or configuration.
		/// </summary>
		/// <param name="ePersist"></param>
		public PagePersistAttribute ( PagePersistEnum ePersist )
		{
			this.ePersist = ePersist;
		}
		/// <summary>
		/// Return the persistence level (medium)
		/// </summary>
		public PagePersistEnum PersistLevel
		{	get { return ePersist; } }
		/// <summary>
		/// Hold the desired persistence level.
		/// </summary>
		protected PagePersistEnum ePersist;
	}
	/// <summary>
	/// Page persistence handling object.  Walks the declared instance or static fields of a
	/// System.Web.UI.Page object and saves or restores field values for fields decorated
	///	with a PagePersistAttribute.  Generally, instance fields are persisted using the Session
	/// property bag and static fields are persisted using the Application property bag.
	/// Note that static fields of ANY class can be persisted using the Application or
	///	web.config files.  Only "page" classes can request persistence of instance fields,
	///	since that mechanism uses the Session object.
	/// </summary>
	/// <remarks>
	/// As the code is currently written, persistence can be applied only to protected fields.  
	///	There are two types of fields: static (class-level or "shared") and object-level. There are three 
	///	persistence media: the Application dictionary, the Session dictionary and the appSesstings 
	///	tag in the project's web.config file.  This leads to six possible combinations of field and media.  In each
	///	case, there is a mapping between the field name you chose and the key used in the relevant dictionary.  
	///	Some media are read-only, others are read-write. Basically, any class can have its static members initialized, but
	///	only subclasses of System.Web.UI.Page can have object-level fields persisted at the session level.
	/// <p>Implications of each combination of field type and storage medium are as follows.</p>
	///	<p>Object field, Session storage: Allowed, save and restore. Object must inherit from System.Web.UI.Page. 
	///	Key is <c>$persist$typeName$fieldName</c>.</p>
	/// <p>Object field, Application storage: Allowed, save and restore. Object must inherit from System.Web.UI.Page.  
	/// Key is <c>$persist$typeName$fieldName</c>.</p>
	/// <p>Object field, Configuration storage: Allowed, but restore only.  
	///	Object must inherit from System.Web.UI.Page.  Key is <c>$persist$fieldName</c>.</p>
	/// <p>Static field, Session storage:  NOT ALLOWED.</p>
	/// <p>Static field, Application storage: Allowed, save and restore. Key is <c>$persist$typeName$fieldName</c>.</p>
	/// <p>Static field, Configuration storage: Allowed, restore only. Key is <c>$persist$typeName$fieldName</c>.</p>
	/// </remarks>
	public class PagePersister : object
	{
		/// <summary>
		/// Construct a PagePersister for a given page.
		/// </summary>
		/// <param name="page">the page object</param>
		public PagePersister ( System.Web.UI.Page page )
		{
			this.page = page;
			this.type = page.GetType().BaseType;
			this.app = page.Application;	
		}

		/// <summary>
		/// Construct a page persister for static fields of the given type.
		/// </summary>
		/// <param name="app">the application state</param>
		/// <param name="type">the Type object for the page type</param>
		public PagePersister ( System.Web.HttpApplicationState app, System.Type type )
		{	
			this.app = app;
			this.type = type;
		}
		/// <summary>
		/// Save instance fields to persistence medium
		/// </summary>
		public void Save ()
		{
			Action(true,true);
		}
		/// <summary>
		/// Restore instance fields from persistence medium
		/// </summary>
		public void Restore ()
		{
			Action(false,true);
		}
		/// <summary>
		/// Save static fields to persistence medium
		/// </summary>
		public void SaveStatic ()
		{
			Action(true,false);
		}
		/// <summary>
		/// Restore static fields from persistence medium
		/// </summary>
		public void RestoreStatic ()
		{
			Action(false,false);
		}
		
		/// <summary>
		/// Restore persistent static fields for the given class (by type)
		/// </summary>
		/// <param name="app">The application state collection for this HttpApplication.</param>
		/// <param name="type">The type of the class whose static members are to be handled.</param>
		public static void RestoreStatic(HttpApplicationState app, Type type)
		{	
			//  Create a temporary object.
			PagePersister persister = new PagePersister(app,type);
			persister.RestoreStatic();			
		}
		/// <summary>
		/// Search for all subclasses of System.Web.UI.Page and restore static fields
		///	to each Page class.
		/// </summary>
		/// <param name="app"></param>
		public static void RestorePageStatics(HttpApplicationState app)
		{
			Assembly asmb = Assembly.GetExecutingAssembly();
			Type[] types = asmb.GetExportedTypes();
			Type tPage = typeof(System.Web.UI.Page);
			foreach ( Type t in types )
			{
				Type tBase = t.BaseType;
				if ( tPage == tBase )
				{
					RestoreStatic(app, t);			
				}
			}
		}
		/// <summary>
		/// Prepare the application's dictionary so that the value of a page's static field
		/// will be set to a particular value when its static fields are restored.
		/// </summary>
		/// <param name="app">The HTTP application object</param>
		/// <param name="typePage">the Type object for the page's type</param>
		/// <param name="sVarName">the name of the field</param>
		/// <param name="objValue">the value to be assigned to the field</param>
		public static void SetPageField(HttpApplicationState app, 
																Type typePage,
																string sVarName, 
																object objValue)
		{
			if ( typePage.BaseType != typeof(System.Web.UI.Page) )
				throw new PagePersistException("invalid type presented for field setting");
			SetPageField(app,typePage.Name,sVarName,objValue);
		}
		/// <summary>
		/// Prepare the application's dictionary so that the value of a page's static field
		/// will be set to a particular value when its static fields are restored.
		/// </summary>
		/// <param name="app">The HTTP application object</param>
		/// <param name="sPageTypeName"></param>
		/// <param name="sVarName">the name of the field</param>
		/// <param name="objValue">the value to be assigned to the field</param>
		public static void SetPageField(HttpApplicationState app, 
			string sPageTypeName,
			string sVarName, 
			object objValue)
		{
			string sVar = VarName(sPageTypeName,sVarName,PagePersistEnum.ePersistApplication);
			app[sVar] = objValue;
		}
		/// <summary>
		/// Set the persistent value of a persisted page field.
		/// </summary>
		/// <param name="session">the session state object</param>
		/// <param name="typePage">the Type of the page class</param>
		/// <param name="sVarName">the declared name of the field</param>
		/// <param name="objValue">the value to be set</param>
		public static void SetPageField(HttpSessionState session, 
			Type typePage,
			string sVarName, 
			object objValue)
		{
			if ( typePage.BaseType != typeof(System.Web.UI.Page) )
				throw new PagePersistException("invalid type presented for preinitialization");
			SetPageField(session,typePage.Name,sVarName,objValue);
		}
		/// <summary>
		/// Set the persistent value of a persisted page field.
		/// </summary>
		/// <param name="session">the session state object</param>
		/// <param name="sPageTypeName">the name of the page class</param>
		/// <param name="sVarName">the declared name of the field</param>
		/// <param name="objValue">the value to be set</param>
		public static void SetPageField(HttpSessionState session, 
			string sPageTypeName,
			string sVarName, 
			object objValue)
		{
			string sVar = VarName(sPageTypeName,sVarName,PagePersistEnum.ePersistApplication);
			session[sVar] = objValue;
		}
		
		/// <summary>
		/// Worker function for persistence: saves or loads static or instance field values.
		/// </summary>
		/// <param name="bSave">true if saving</param>
		/// <param name="bInstance">true if instance</param>
		protected void Action(bool bSave, bool bInstance)
		{
			BindingFlags bflags = bInstance 
								? bindFlagsInstance 
								: bindFlagsStatic;
								
			bool bAppLocked = false;
			try
			{
				MemberInfo[] mbrinfos = type.GetMembers( bflags );
				
				string sTypeName = type.Name;
				
				for ( int imbr = 0; imbr < mbrinfos.Length; ++imbr)
				{
					MemberInfo mbrinfo = mbrinfos[imbr];
					switch ( mbrinfo.MemberType )
					{
						case MemberTypes.Field:
						{
							FieldInfo finfo = (FieldInfo) mbrinfo;
							object[] attrs = finfo.GetCustomAttributes(typeof(PagePersistAttribute),false);
							for ( int iattr = 0; iattr < attrs.Length; ++iattr )
							{
								PagePersistAttribute pattr = (PagePersistAttribute)attrs[iattr];
								PagePersistEnum ePersist = pattr.PersistLevel;
								string sName = VarName(finfo,ePersist);

								switch ( ePersist )
								{	
									case PagePersistEnum.ePersistSession:
									{	
										if ( ! bInstance )
											break;

										if ( page == null )
											throw new PagePersistException("missing page object reference for instance field");
										object oSess = null;
										if ( bSave )
											oSess = finfo.GetValue(page);
										else
											finfo.SetValue(page, page.Session[sName]);
										// Always update the session collection; set a new value or clear the old one.
										// This avoids having an silent object reference in the Session collection
										page.Session[sName] = oSess;
										break;
									}
										
									case PagePersistEnum.ePersistApplication:
									{
										//  N.B.: according to the docs, the object reference on "GetValue" and
										//	"SetValue" is IGNORED for static fields.										
										if ( bInstance && page == null )
											throw new PagePersistException("missing page object reference for instance field (application)");
										
										//  If this is the first usage of the Application object, lock it
										if ( ! bAppLocked )
										{
											bAppLocked = true;	
											app.Lock();
										}
										object oApp = null;
										if ( bSave )
											oApp = finfo.GetValue(page);
										else
											finfo.SetValue(page, app[sName]);
										// See note about about always updating the session collection
										app[sName] = oApp;
										break;
									}
									
									case PagePersistEnum.ePersistWebConfig:
									{	// Ignore attempts to save info to web.config.
										// Also, don't overwrite info with nulls (leave default values).
										if ( ! bSave )
										{
											//  N.B.: according to the docs, the object reference on "GetValue" and
											//	"SetValue" is IGNORED for static fields.										
											if ( bInstance && page == null )
												throw new PagePersistException("missing page object reference for instance field (configuration)");
											string sCfg = ConfigurationSettings.AppSettings[sName];
											if ( sCfg != null )
												finfo.SetValue(page, sCfg);
										}
										break;
									}
									default:
										break;
								}
							}
							break;
						}
						
						default:
							break;
					}
				}
			}
			catch (Exception ex)
			{
				throw new PagePersistException("page persistence failure",ex);
			}
			finally
			{
				if ( bAppLocked )
				{
					app.UnLock();
				}
			}
		}
		/// <summary>
		/// Perform page-level persistence for the newly-created or about-to-be-released page
		/// </summary>
		/// <param name="app">the HttpApplication</param>
		/// <param name="bRestore">true if restoring the page fields</param>
		public static void PersistPage(HttpApplication app, bool bRestore)
		{
			//  Access the page object that will handle this request.
			System.Web.UI.Page page = app.Context.Handler as System.Web.UI.Page;
			if ( page == null )
				return;		//  For now, we only handle standard ASPX page subclasses.
				
			//  Create a persister helper for operating on this page
			PagePersister persist = new PagePersister(page);
			
			//	Unfortunately, the page's IsPostBack property is not yet set at this stage, so I 
			//	attempt to deduce it by checking whether this is a new session or the request is a GET.
			string sMethod = app.Context.Request.HttpMethod.Trim().ToUpper();
			bool bInitialRequest = app.Session.IsNewSession || sMethod != "POST";
			if ( bRestore )
			{
				if ( ! bInitialRequest )
					persist.Restore();
			}
			else
			{
				persist.Save();
			}
		}

		/// <summary>
		/// Clean up any persisted field values from this session's state collection.
		///	That is, call IDisposable.Dispose() on any objects that support it.
		/// </summary>
		/// <param name="session"></param>
		public static void SessionEnd (System.Web.SessionState.HttpSessionState session)
		{
			foreach ( string sKey in session.Keys )
			{	//  See if this key begins with the session persistence prefix
				if ( sKey.IndexOf(sPrefixSession,0) == 0 )
				{
					object oValue = session[sKey];
					try
					{
						IDisposable idisp = (IDisposable)oValue;
						idisp.Dispose();
					}
					catch
					{
					}
				}				
			}
		}
		/// <summary>
		/// Generate a string name to use as a key for a persistence data collection.
		/// For Session data, the field key is "$persist$classTypeName$fieldName".
		/// For Application data, the field key is "$persist$classTypeName$fieldName".
		/// For Web Config data, the field key is "$persist$fieldName".
		/// </summary>
		/// <param name="finfo">The field's type information</param>
		/// <param name="penum">The persistence type (app, session or web.config)</param>
		/// <returns></returns>
		protected string VarName(FieldInfo finfo, PagePersistEnum penum)
		{
			return VarName(type.Name,finfo.Name,penum);
		}
		/// <summary>
		/// Generate a string name to use as a key for a persistence data collection.
		/// For Session data, the field key is "$persist$classTypeName$fieldName".
		/// For Application data, the field key is "$persist$classTypeName$fieldName".
		/// For Web Config data, the field key is "$persist$fieldName".
		/// </summary>
		/// <param name="sTypeName">name of the object/page type</param>
		/// <param name="sVarName">name of the field/variable</param>
		/// <param name="penum">type of persistence storage</param>
		/// <returns></returns>
		protected static string VarName(string sTypeName, string sVarName, PagePersistEnum penum)
		{
			string sName;
			string sPrefix = "";

			switch ( penum )
			{
				case PagePersistEnum.ePersistSession:
					sPrefix = sPrefixApplication;
					goto default;
				case PagePersistEnum.ePersistApplication:
					sPrefix = sPrefixSession;
					goto default;
				default:
					sName = string.Format("{0}{1}{2}{3}{4}",
						sPrefix,
						sPrefixDelimiter,
						sTypeName,
						sPrefixDelimiter,
						sVarName);
					break;

				case PagePersistEnum.ePersistWebConfig:
					sName = string.Format("{0}{1}{2}",
						sPrefixConfiguration,
						sPrefixDelimiter,
						sVarName);
					break;					
			}
			return sName;
		}
		/// <summary>
		/// Delimiter character used between key name components
		/// </summary>
		protected static string sPrefixDelimiter = "$";
		/// <summary>
		/// String prefix to use for session fields (along with type and field name)
		/// </summary>
		protected static string sPrefixSession = "$persist";
		/// <summary>
		/// String prefix to use for session fields (along with type and field name)
		/// </summary>
		protected static string sPrefixApplication = "$persist";
		/// <summary>
		/// String prefix to use for web.config fields (along with field name)
		/// </summary>
		protected static string sPrefixConfiguration = "$persist";
		/// <summary>
		/// The web application for this operation
		/// </summary>
		protected System.Web.HttpApplicationState app = null;		
		/// <summary>
		/// The page, if any, for this operation 
		/// </summary>
		protected System.Web.UI.Page page = null;
		/// <summary>
		/// The type of page or object being operated on.
		/// </summary>
		protected System.Type type = null;
		
		/// <summary>
		/// The type information binding flags used for walking the instance fields. 
		/// </summary>
		protected static readonly BindingFlags bindFlagsInstance = 
									  BindingFlags.Instance 
									| BindingFlags.Public 
									| BindingFlags.NonPublic 
									| BindingFlags.DeclaredOnly;
		/// <summary>
		/// The type information binding flags used for walking the static fields
		/// </summary>	
		protected static readonly BindingFlags bindFlagsStatic = 
									  BindingFlags.Static 
									| BindingFlags.Public 
									| BindingFlags.NonPublic 
									| BindingFlags.DeclaredOnly;
	}
}
